/* Array Processor interface                                                 */

/*--- Microsoft C 5.0 compiler include files                                 */

#include <conio.h>
#include <dos.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*--- other includes                                                         */

#include "driver.h"

#define ARRAY
#include "array.h"

#include "data.h"
#include "extmath.h"
#include "graphics.h"
#include "screen.h"
#include "serial.h"


/*---------------------------------------------------------------------------*/
/* Error messages if the array fails in flight                               */
/* These get shown after the graphics screen is destroyed                    */

char nochain[] =    "*** Array didn't return chain signal after computation";
char badengine1[] = "*** Bad engine address during initialization";
char badengine2[] = "*** No chain return during engine address assignment";
char badROMs[]    = "*** Array has different ROM versions";

/*---------------------------------------------------------------------------*/
/* Array interface structures                                                */
/* Remember that the final step must be to reverse the byte order!           */
/* The array stores all numbers with high bytes first!                       */

#define BROADCAST 0xff                 /* all array elements notice this one */

#define BCASTBIT 0x80                  /*  ... if this opcode bit is set     */

#define OP_CANCEL   0                  /* opcode values                      */
#define OP_LOADCOM  1
#define OP_LOADGO   2
#define OP_ROMver   6
#define OP_IDstr    7

typedef struct {                       /* data common to whole array         */
  unsigned int maxiter;
  unsigned int Rsteps;
  unsigned int Isteps;
  unsigned int reserved0;
  EXTNUM       Rstep;
  EXTNUM       Istep;
} OP1DATA;

typedef struct {                       /* data unique to each processor      */
  EXTNUM C_R;
  EXTNUM C_I;
  EXTNUM Z_R;
  EXTNUM Z_I;
} OP2DATA;


/*---------------------------------------------------------------------------*/
/* Draw one pass worth of dots from the array                                */

void DrawPass(void) {

  int enginenum;

  ExtractCalc();

  for (enginenum = 0; enginenum < NumEngines; enginenum++) {

     if (Faking) {                               /* fake some answers?       */
        if (TestPattern == 0) {
           EngineValue[enginenum] = CompMandel(ThisDot.cplx);
        }
     }

     IterTotal += (long unsigned int) EngineValue[enginenum];

     if (IterLimit == EngineValue[enginenum]) {
        ThisDot.scrn.color = 0;
        InMSet++;
     }
     else {
        ThisDot.scrn.color =
            min(EngineValue[enginenum] >> ColorShift, MaxColor);
        if (!ThisDot.scrn.color)
           ThisDot.scrn.color = 1;
     }

     SetPel(ThisDot.scrn.x,ThisDot.scrn.y,ThisDot.scrn.color);
     ThisDot.scrn.x++;
     ThisDot.cplx.r += PelSize.r;
  }

/*--- wrap at end of each row                                                */

  if (!(PassCounter % PassesPerRow)) {
     ThisDot.scrn.x = ResultCenter.scrn.x - (ResultSize.scrn.x/2);
     ThisDot.cplx.r = ResultCenter.cplx.r - (ResultSize.cplx.r/2);

     ThisDot.scrn.y++;
     ThisDot.cplx.i += PelSize.i;

     if (PassCounter >= NumPasses) {
        ThisDot.scrn.y = ResultCenter.scrn.y - (ResultSize.scrn.y/2);
        ThisDot.cplx.i = ResultCenter.cplx.i - (ResultSize.cplx.i/2);
     }
  }


} /* DrawPass */


/*---------------------------------------------------------------------------*/
/* Set up for a new round of calculations                                    */
/* This abandons any calculations in progress                                */
/* The array is halted, reloaded, and started                                */
/* The calculations fill the result area                                     */

void InitialCalc(void) {

  int op;
  OP1DATA d1;
  OP2DATA d2;
  int enginenum;

  if (EngineState == unknown)                    /* "this should not be"     */
     SetupArray();

  SplitNew();
  printf("Calculation setup...\n");
  SplitScreen(1,NORMAL,INSTANT);

/*--- refigure the calculations required based on the result size            */

  PassesPerRow = ResultSize.scrn.x / NumEngines;
  NumPasses = ResultSize.scrn.y * PassesPerRow;


/*--- cancel the current computations                                        */

  SendAddress(BROADCAST);
  op = OP_CANCEL | BCASTBIT;

  SendData(op);


/*--- tell the array to start over                                           */

  SendAddress(BROADCAST);
  op = OP_LOADCOM | BCASTBIT;

  SendData(op);

/*--- now tell it what the calculation range is all about                    */
/*    for simplicity, we calculate a row of pels the size of the array       */
/*     with each pass... makes the displaying easier                         */

  printf("  global ");

  d1.maxiter = _rotr(IterLimit,8);
  d1.Rsteps  = _rotr(((unsigned int) PassesPerRow),8);
  d1.Isteps  = _rotr(ResultSize.scrn.y,8);
  d1.reserved0 = _rotr(TestPattern,8);           /* debugging value!         */
  CvtDbl2FullExt(d1.Rstep,PelSize.r*(double)NumEngines);
  SwapExt(d1.Rstep);
  CvtDbl2FullExt(d1.Istep,PelSize.i);
  SwapExt(d1.Istep);

  SendBlock((char *)&d1.maxiter,sizeof(d1));


/*--- figure where the first dot lies in the complex plane (and screen)      */

  ThisDot.scrn.x = ResultCenter.scrn.x - (ResultSize.scrn.x/2);
  ThisDot.cplx.r = ResultCenter.cplx.r - (ResultSize.cplx.r/2);

  ThisDot.scrn.y = ResultCenter.scrn.y - (ResultSize.scrn.y/2);
  ThisDot.cplx.i = ResultCenter.cplx.i - (ResultSize.cplx.i/2);

  ThisDot.scrn.color = 0;


/*--- now assign each processor to part of the task                          */
/*    each will start when it's got all the data it needs                    */

  op = OP_LOADGO;

  time(&StartTime);                              /* starts with first send   */
  ElapsedTime = 0.0;
  IterTotal = 0L;                                /* with no iterations       */
  InMSet = 0L;                                   /* and nothing in the set   */

  printf("and per-engine");

  for (enginenum = 0; enginenum < NumEngines; enginenum++) {

     SendAddress(enginenum);
     SendData(op);

     CvtDbl2FullExt(d2.C_R,ThisDot.cplx.r + ((double) enginenum) * PelSize.r);
     SwapExt(d2.C_R);
     CvtDbl2FullExt(d2.C_I,ThisDot.cplx.i);
     SwapExt(d2.C_I);
     memset(d2.Z_R,'\0',NUMLEN);
     memset(d2.Z_I,'\0',NUMLEN);

     SendBlock((char *)d2.C_R,sizeof(d2));
  }

  EngineState = running;                         /* they're off!             */
  PassCounter = 0;
  Changed = 1;

  SplitScreen(0,NORMAL,INSTANT);

}


/*---------------------------------------------------------------------------*/
/* Extract the most recent array results                                     */
/* This keeps track of the number of passes and clears the state at the end  */
/* If the array doesn't blip the chain in line, we make an emergency exit    */
/* This can't be interrupted by much else because the array doesn't wait!    */

void ExtractCalc(void) {

  int enginenum;
  int *presult;
  time_t start,now;
  int chaintest;

  if (!Faking) {
     presult = EngineValue;

     enginenum = 0;
     if (IterLimit <= 255) {
        PauseScroll(1);                          /* stop scrolling           */
        PulseChain();
        do {                                     /* expect one byte          */
           while (!(DR & inp(CommPort+LSR)));
           *presult = inp(CommPort+RB);
           presult++;
        } while (NumEngines > ++enginenum);
        PauseScroll(0);                          /* resume scrolling         */
     }
     else {
        PauseScroll(1);                          /* stop scrolling           */
        PulseChain();
        do {                                     /* expect two bytes         */
           while (!(DR & inp(CommPort+LSR)));
           *presult = inp(CommPort+RB) << 8;     /*   high byte first        */
           while (!(DR & inp(CommPort+LSR)));
           *presult |= inp(CommPort+RB);         /*   then low byte          */
           presult++;
        } while (NumEngines > ++enginenum);
        PauseScroll(0);                          /* resume scrolling         */
     }

     if (CheckChain()) {
        time(&start);
        do {
           chaintest = CheckChain();
        } while (chaintest && (ARRAYTIMEOUT > difftime(time(&now),start)));
        if (chaintest) {
           SplitNew();
           printf("*** Timeout waiting for chain return!");
           SplitScreen(1,FAST,INSTANT);
           ErrorMessage = nochain;
           longjmp(ErrorEnv,1);
        }
     }
  }
  else {                                         /* fake an easy result      */
     for (enginenum = 0; enginenum < NumEngines; enginenum++) {
        EngineValue[enginenum] = enginenum;
     }

  }

/*--- tick the pass counter and decide if we're done with the array yet      */

  PassCounter++;
  if (PassCounter >= NumPasses) {
     DrawRun = 0;                      /* finished, stop drawing loop        */
     EngineState = idle;
     EngineRuns++;
     time(&FinishTime);
     ElapsedTime = difftime(FinishTime,StartTime);
  }

  return;

}

/*---------------------------------------------------------------------------*/
/* Fake a Mandelbrot set calculation using the extended math routines        */
/* This is a rather length process...                                        */
/* Faking = 1 uses floating point math with 80287 if installed...            */
/*        = 2 same as 2 with "float" instead of "double" results             */
/*        = 3 uses C version of integer code                                 */

unsigned int CompMandel(COMPLEX zpt) {
  unsigned int iterctr;
  EXTNUM cr,ci;                        /* current iteration values           */
  EXTNUM r,i;
  EXTNUM rr,ii;                        /* square of iteration values         */
  EXTNUM work;

  COMPLEX dc;                          /* ditto, for floating point          */
  COMPLEX dz;
  COMPLEX dzz;

typedef struct {
  float r;
  float i;
} FCOMPLEX;

  FCOMPLEX fc;                         /* ditto, for low-precision calcs     */
  FCOMPLEX fz;
  FCOMPLEX fzz;

  switch (Faking) {                              /* set up variables         */
     case (1) :                                  /* double-precision         */
        dc = zpt;
        dz.r =  0.0;
        dz.i =  0.0;
        dzz.r = 0.0;
        dzz.i = 0.0;
        break;
     case (2) :                                  /* single-precision         */
        fc.r =  (float) zpt.r;
        fc.i =  (float) zpt.i;
        fz.r =  (float) 0.0;
        fz.i =  (float) 0.0;
        fzz.r = (float) 0.0;
        fzz.i = (float) 0.0;
        break;
     case (3) :                                  /* fixed point              */
        CvtDbl2Ext(r,0.0);
        CvtDbl2Ext(i,0.0);
        CvtDbl2Ext(rr,0.0);
        CvtDbl2Ext(ii,0.0);
        CvtDbl2FullExt(cr,zpt.r);
        CvtDbl2FullExt(ci,zpt.i);
        break;
  }

  switch (Faking) {
     case (1) :
        for (iterctr=1; iterctr < IterLimit; iterctr++) {
           dz.i = (2.0 * dz.i * dz.r) + dc.i;
           dz.r = dzz.r - dzz.i + dc.r;
           dzz.r = dz.r * dz.r;
           dzz.i = dz.i * dz.i;
           if ((dzz.r >= 4.0) || (dzz.i >= 4.0) ||
               ((dzz.r + dzz.i) >= 4.0)) {
              goto endloop;
           }
        }
        break;

     case (2) :
        for (iterctr=1; iterctr < IterLimit; iterctr++) {
           fz.i = (float) ((2.0 * fz.i * fz.r) + fc.i);
           fz.r = (float) (fzz.r - fzz.i + fc.r);
           fzz.r = (float) (fz.r * fz.r);
           fzz.i = (float) (fz.i * fz.i);
           if (((double) fzz.r >= 4.0) || ((double) fzz.i >= 4.0) ||
               ((double) (fzz.r + fzz.i) >= 4.0)) {
              goto endloop;
           }
        }
        break;

     case (3) :
        for (iterctr=1; iterctr < IterLimit; iterctr++) {

           memcpy(work,MultiplyExt(r,i),NUMLEN); /* form 2*I*R + CI          */
           Mpy2Ext(work);
           memcpy(ExtAccum,work,NUMLEN);
           memcpy(i,AddExt(ci),NUMLEN);          /*  ... save as new I       */

           NegExt(ExtAccum,ii);                  /* form RR-II + CR          */
           AddExt(rr);
           memcpy(r,AddExt(cr),NUMLEN);          /*  ... save as new R       */

           memcpy(rr,MultiplyExt(r,r),NUMLEN);   /* square R and I           */
           memcpy(ii,MultiplyExt(i,i),NUMLEN);

           if ((0x40 <= *(rr+NUMLEN-1)) ||       /* done if either >= 4.0    */
               (0x40 <= *(ii+NUMLEN-1))) {
              goto endloop;
           }
           else {
              memcpy(ExtAccum,rr,NUMLEN);
              if (0x40 <= *(AddExt(ii)+NUMLEN-1)) {        /* or sum >= 4.0  */
                 goto endloop;
              }
           }
        }
        break;

  }

endloop :
  return iterctr;

}


/*---------------------------------------------------------------------------*/
/* Initialize the array with a hard reset                                    */
/* Assigns engine addresses from scratch                                     */
/* Checks to make sure that all the ROM versions are the same                */

void SetupArray(void) {

  int engineout;
  int enginein;
  int op;
  unsigned int ROMbase,ROMcode;                  /* array ROM versions       */
  char IDstring[80];
  char *IDchar;

  time_t start,now;
  double delta;
  int loops;

  SplitNew();
  printf("Assigning Engine Addresses...\n");
  SplitScreen(1,FAST,INSTANT);

  if (!Faking) {

     engineout = 0;
     NumEngines = 0;

     PauseScroll(1);                             /* stop scrolling           */
     PulseChain();                               /* start chain going        */

     do {
        SendAddress(engineout);                  /* send engine number       */
        time(&start);
        loops = 0;

        do {                                     /* wait for a response      */
           enginein = ReadPort();
           loops++;
        } while ((enginein == -1) &&
               (ARRAYTIMEOUT > (delta = difftime(time(&now),start))));

        if (enginein != -1) {                    /* something from array     */
           if (0x00ff == (0x00ff & (enginein + engineout)))
              engineout++;                       /* good checksum!           */
           else {
              PauseScroll(0);
              SplitNew();
              printf("Bad address return at engine %i\n"
                     "  got %2.2X, expected %2.2X",
                       enginein,engineout,(0x00ff & ~engineout));
              ErrorMessage = badengine1;
              longjmp(ErrorEnv,-1);
           }
        }
        else {                                   /* found last engine?       */
           if (CheckChain()) {
              PauseScroll(0);
              SplitNew();
              printf("No chain return at engine %i\n"
                     "  after timeout",engineout);
              ErrorMessage = badengine2;
              longjmp(ErrorEnv,-1);
           }
           else {
              NumEngines = engineout;            /*  yes!  count = index + 1 */
           }
        }

     } while (NumEngines == 0);

     PauseScroll(0);                              /* resume scrolling         */
  }
  else                                           /* a faked response here..  */
     NumEngines = 64;

  printf("  there are %i engines",NumEngines);


/*--- check the ROM version codes                                            */

  SplitScreen(0,NORMAL,TEASER);

  SplitNew();
  printf("Array ROMs: ");
  SplitScreen(1,FAST,INSTANT);

  op = OP_ROMver;

  ROMbase = inp(CommPort+RB);                    /* clear data ready if on   */

  SendAddress(0);                                /* get first ROM code       */
  SendData(op);

  if (!Faking) {
     while (!(DR & inp(CommPort+LSR)));          /* get two bytes back       */
     ROMbase = inp(CommPort+RB) << 8;            /*   high byte first        */
     while (!(DR & inp(CommPort+LSR)));
     ROMbase |= inp(CommPort+RB);                /*   then low byte          */
  }
  else
     ROMbase = 0x0000;

  printf("V %X.%2.2X\n",ROMbase>>8,ROMbase&0x00ff);

  if (!Faking) {
     for (engineout = 1; engineout < NumEngines; engineout++) {

        SendAddress(engineout);
        SendData(op);

        while (!(DR & inp(CommPort+LSR)));          /* get two bytes back       */
        ROMcode = inp(CommPort+RB) << 8;            /*   high byte first        */
        while (!(DR & inp(CommPort+LSR)));
        ROMcode |= inp(CommPort+RB);                /*   then low byte          */

        if (ROMcode != ROMbase) {                   /* matches?                 */
           printf("\n  Oops! ROM V %X.%2.2X at engine %i",
                   ROMcode>>8,ROMcode&0x00ff,engineout);
           ErrorMessage = badROMs;
           longjmp(ErrorEnv,-1);

        }

     }
  }

/*--- display the array copyright message                                    */

  op = OP_IDstr;

  ROMbase = inp(CommPort+RB);                    /* clear data ready if on   */

  SendAddress(0);                                /* get first ROM code       */
  SendData(op);

  if (!Faking) {
     IDchar = IDstring - 1;
     do {
        while (!(DR & inp(CommPort+LSR)));
        *(++IDchar) = (char) inp(CommPort+RB);
     } while (*IDchar != '\0');
  }
  else
     strcpy(IDstring,"Simulated array!");

  printf("  %s",IDstring);
  SplitScreen(0,NORMAL,READABLE);


/*--- now set up the rest of the variables                                   */

  EngineRuns = 0;

  EngineState = idle;


  return;

}

